"""Data class helper for PDP/PDN Context.

The Packet Data Protocol or Packet Data Network Context is the logical
connection between a mobile device and a mobile network that allows exchange of
packets.
"""
from dataclasses import dataclass

from pynbntnmodem.constants import PdpType


@dataclass
class PdpContext:
    """Attributes of a NB-NTN Packet Data Protocol context/definition."""
    id: int = 1   # context ID
    pdp_type: PdpType = PdpType.IP
    apn: str = ''
    ip: 'str|None' = ''   # the IP address if type is IP and attached



"""Data class helper for UDP socket status for NB-NTN."""

from dataclasses import dataclass


@dataclass
class SocketStatus:
    """Metadata for a UDP socket including state and IP address.
    
    Attributes:
        active (bool): Indicator whether the socket is active.
        ip_address (str): The IP address assigned by the network.
    """
    active: bool = False
    ip_address: str = ''



"""Data class helper for Messages.

Provides an abstraction allowing for distinction of payload from IP overhead,
and determination of packet size.
"""

import ipaddress
from dataclasses import dataclass
from typing import Optional, Union

from pynbntnmodem.constants import PdpType

__all__ = [ 'BaseMessage', 'MoMessage', 'MtMessage' ]

@dataclass
class BaseMessage:
    payload: bytes
    transport: PdpType
    id: Optional[Union[int, str]] = None
    
    @property
    def size(self) -> int:
        if self.transport is None or self.payload is None:
            return 0
        if self.transport == PdpType.NON_IP:
            return len(self.payload)
        if self.transport == PdpType.IP:
            return len(self.payload) + 20 + 8
        if self.transport == PdpType.IPV6:
            return len(self.payload) + 40 + 8
        ip = None
        if hasattr(self, 'dst_ip'):
            ip = getattr(self, 'dst_ip')
        elif hasattr(self, 'src_ip'):
            ip = getattr(self, 'src_ip')
        if not ip:
            return 0
        try:
            parsed_ip = ipaddress.ip_address(ip)
            if isinstance(parsed_ip, ipaddress.IPv4Address):
                return len(self.payload) + 20 + 8
            return len(self.payload) + 40 + 8
        except ValueError:
            return 0


@dataclass
class MoMessage(BaseMessage):
    """Metadata for Mobile-Terminated message including payload and source"""
    dst_ip: Optional[str] = None
    dst_port: Optional[int] = None
    

@dataclass
class MtMessage(BaseMessage):
    """Metadata for Mobile-Terminated message including payload and source"""
    src_ip: Optional[str] = None
    src_port: Optional[int] = None



"""This is a placeholder"""
import importlib.util
import logging
import os
import tempfile
import shutil
import subprocess
from typing import Type

from pynbntnmodem.modem import NbntnBaseModem

__all__ = ['clone_and_load_modem_classes', 'load_modem_class']

_log = logging.getLogger(__name__)


def clone_and_load_modem_classes(repo_urls: 'list[str]',
                                 branch: str = 'main',
                                 download_path: str = '',
                                 ) -> dict[str, Type[NbntnBaseModem]]:
    """
    Clone multiple Git repositories and load subclasses of NbntnModem.

    Args:
        repo_urls (list[str]): A list of Git repository URLs.
        branch (str): The branch to clone. Defaults to 'main'.

    Returns:
        dict[str, Type[NbntnModem]]: A dictionary of modem class names and their corresponding classes.
    """
    modem_classes = {}
    
    # Create a temporary directory to clone repositories
    with tempfile.TemporaryDirectory() as temp_dir:
        for repo_url in repo_urls:
            repo_name = repo_url.split("/")[-1].replace(".git", "")
            repo_path = os.path.join(temp_dir, repo_name)
            _log.debug("Cloning git repository into %s...", repo_path)
            result = subprocess.run(
                ["git", "clone", "--branch", branch, repo_url, repo_path],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            if result.returncode != 0:
                _log.error("Failed to clone repository %s: %s",
                           repo_url, result.stderr)
                continue

            _log.debug("Git repository %s cloned successfully.", repo_name)

            # Find Python files in the repository and load modem classes
            for root, _, files in os.walk(repo_path):
                for file in files:
                    if file.endswith(".py"):
                        file_path = os.path.join(root, file)
                        class_def = load_modem_class(file_path)
                        if class_def:
                            modem_classes[file.replace('.py', '')] = class_def
                            if download_path and os.path.isdir(download_path):
                                dest_path = os.path.join(download_path, file)
                                shutil.copy(file_path, dest_path)
                                _log.debug('Copied %s to %s', file, dest_path)

    return modem_classes


def load_modem_class(file_path: str) -> Type[NbntnBaseModem] | None:
    """
    Load a Python file and return the modem class if it subclasses NbntnModem.

    Args:
        file_path (str): Path to the Python file.

    Returns:
        Type[NbntnBaseModem] | None: The modem class if valid, else None.
    """
    module_name = os.path.splitext(os.path.basename(file_path))[0]
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)

    try:
        spec.loader.exec_module(module)
    except Exception as e:
        print(f"Error loading {file_path}: {e}")
        return None

    # Look for subclasses of NbntnModem
    for attr_name in dir(module):
        attr = getattr(module, attr_name)
        if (isinstance(attr, type) and issubclass(attr, NbntnBaseModem) and
            attr is not NbntnBaseModem):
            return attr

    return None
